#include <iostream>
#include <fstream>
#include <string>
#include <vector>
#include <cmath>
#include <cstdint>
#include <stdexcept>
#include <algorithm>
#include <iomanip>

#include <Eigen/Core>


// used for the input proton spectrum and the related metadata
typedef struct {
	Eigen::VectorXd bins;
	uint16_t min = 0;
	uint16_t max = 0;	
	uint32_t time = 0;
	double crystal_area = 3.141592653589793238463 * 2.25e-2 * 2.25e-2;
	double step = 0.1;
} proton_spectrum;


// reads proton spectrum from file
proton_spectrum read_proton_spectrum(std::istream &spectrum_stream)
{
	if(spectrum_stream.fail())
		throw std::invalid_argument("failed to read proton spectrum file");

	proton_spectrum spectrum;
	std::vector<double> elements;
	
	while(spectrum_stream.good()) {
		if((spectrum_stream >> std::ws).peek()
					   	== std::char_traits<char>::to_int_type('#')) {
			std::string control_word;
			spectrum_stream >> control_word;
			if(control_word == "#min") spectrum_stream >> spectrum.min;
			else if(control_word == "#max") spectrum_stream >> spectrum.max;
			else if(control_word == "#time") spectrum_stream >> spectrum.time;
			else if(control_word == "#crystal_area") spectrum_stream >> spectrum.crystal_area;
			else if(control_word == "#step") spectrum_stream >> spectrum.step;
			else throw std::runtime_error("unknown control token: " + control_word);	
		} else {
			double element;
			spectrum_stream >> element;
			if(!spectrum_stream.fail())
				elements.push_back(element);
		}
	}	
	
	if(!spectrum_stream.eof())
		throw std::runtime_error("unable to read from spectrum file");	

	if(elements.size() != spectrum.max - spectrum.min + 1)
		throw std::runtime_error("spectrum size and #min and #max mismatched");

	spectrum.bins = Eigen::VectorXd::Zero(elements.size());	
	for(size_t i = 0; i < elements.size() ; ++i)
		spectrum.bins(i) = elements[i];
	
	return spectrum;
}


// read the response matrix from file
Eigen::ArrayXXd read_response_matrix(std::istream &matrix_stream)
{
	if(matrix_stream.fail())
		throw std::invalid_argument("failed to read response matrix file");

	std::vector<double> matrix_elements;

	while(matrix_stream.good()) {
		double element = 0.;
		matrix_stream >> element;
		if(!matrix_stream.fail())
			matrix_elements.push_back(element);
	}
	
	if(!matrix_stream.eof())
		throw std::runtime_error("unable to read from matrix file");	

	uint16_t size = std::sqrt(matrix_elements.size());
	if(size * size != matrix_elements.size() || size < 1)
		throw std::invalid_argument("non-square response matrix");

	Eigen::ArrayXXd response_matrix = Eigen::ArrayXXd::Zero(size, size);
	for(uint16_t i = 0; i < size; i++)
		for(uint16_t j = 0; j < size; j++)
			response_matrix(i, j) = matrix_elements[i*size + j];

	return response_matrix;
}


// Take the response function and proton spectrum and estimate
// the original neutron spectrum.
// 
// First, the spectrum and the response matrix are aligned by energy
// and cut down to size of the smaller one and normalized.
Eigen::ArrayXd unfold(Eigen::ArrayXXd response_function,
			           Eigen::VectorXd proton_spectrum,
					   size_t min_band, size_t max_band,
					   size_t iterations, double precision=0.)
{
	size_t top_limit = std::min(max_band, (size_t) response_function.rows() - 1);
	std::cerr << "using " << top_limit - min_band + 1 << " bands" << std::endl;

	size_t first = min_band - 1, count = top_limit - min_band + 1;
	// align and match the sizes of the response function and the spectrum
	Eigen::ArrayXXd response_cutout = response_function.block(first, first, count, count);
	Eigen::ArrayXd norms = response_cutout.colwise().sum();
	Eigen::MatrixXd response_normed = response_cutout.rowwise() / norms.transpose();

	// resize and normalize the proton spectrum
	Eigen::ArrayXd proton_cutout = proton_spectrum.head(count);
	Eigen::ArrayXd proton_normed = proton_cutout / proton_cutout.sum();

	// uniform initialization
	Eigen::VectorXd result = Eigen::VectorXd::Ones(count) / count;

	// one iteration of the E-M algorithm
	// the matrix library uses the other direction for vectors, the order
	// of operations needs to be different than in the Python implementation
	auto iterate = [&]() -> Eigen::ArrayXd {
		Eigen::ArrayXd e = response_normed * result;
		Eigen::ArrayXd f_tmp = response_normed.transpose() * (proton_normed / e).matrix();
		return f_tmp * result.array();
	};

	// the relative difference between the old and new iterations
	auto difference = [](Eigen::VectorXd &a, Eigen::VectorXd &b) -> double {
		return (a - b).norm() / a.norm();
	};

	std::cerr << "starting E-M" << std::endl;

	for(size_t i = 0; i < iterations; i++) {
		Eigen::VectorXd new_result = iterate();

		double diff = difference(result, new_result);
		result = new_result;
		std::cerr << "iteration " << i + 1 << ": diff " << diff << std::endl;

		if(diff < precision) {
			std::cerr << "precision reached, terminating" << std::endl;
		    break;	
		}
	}

	// scale back to absolute values
	return result.array() * proton_cutout.sum() / norms;
}


int main(int argc, char *argv[])
{
	try {
        uint32_t iterations = 10;
		double limit = 0.;
		if(argc < 3 || argc > 5) {
			std::cerr << "usage: unfold MATRIX PROTON_SPECTRUM"
				      << " [MAX_ITERATIONS [PRECISION]]" << std::endl;
			return 1;
		}
		
		std::ifstream matrix_stream(argv[1]);		
		auto response_matrix = read_response_matrix(matrix_stream);
		std::cerr << "response matrix loaded" << std::endl;

		proton_spectrum spectrum;
		if(std::string(argv[2]) == "-") {
			spectrum = read_proton_spectrum(std::cin);
		} else {
            std::ifstream spectrum_stream(argv[2]);
			spectrum = read_proton_spectrum(spectrum_stream);
		}
		
		if(argc > 3) {
			int iterations_raw = std::stoi(argv[3]);

			if(iterations < 0)
				throw std::runtime_error("bad value for MAX_ITERATIONS");
			iterations = iterations_raw;

			if(argc > 4) {
				double limit_raw = std::stod(argv[4]);
				if(limit_raw < 0.)
					throw std::runtime_error("bad value for LIMIT");
				limit = limit_raw;
			}
		}
		
		std::cerr << "proton spectrum loaded" << std::endl;
		std::cerr << "response function size: "
			   	<< response_matrix.cols() << std::endl;
		std::cerr << "proton spectrum size: "
			   	<< spectrum.max - spectrum.min + 1 << std::endl;
		std::cerr << "crystal area: "
			   	<< spectrum.crystal_area << " m^-2" << std::endl;
		std::cerr << "step: " << spectrum.step << std::endl;
		std::cerr << "lowest energy: "
			   	<< spectrum.step * (spectrum.min - 0.5) << " MeV" << std::endl;
		std::cerr << "highest energy: "
			   	<< spectrum.step * (spectrum.max - 0.5) << " MeV" << std::endl;
		auto res = unfold(response_matrix, spectrum.bins,
						  spectrum.min, spectrum.max,
						  iterations, limit);

		auto energies = Eigen::ArrayXd::LinSpaced(
						    spectrum.max - spectrum.min + 1,
                            (spectrum.min - 0.5) * spectrum.step,
					        (spectrum.max - 0.5) * spectrum.step);

		auto normed = res / (spectrum.step * spectrum.step * spectrum.crystal_area * spectrum.time);
	
		std::cout << std::setprecision(15);
		for(size_t i = 0; i < energies.size(); i++)
			std::cout << energies[i] << "\t" << normed[i] << std::endl;

		return 0;
	} catch (const std::exception &ex) {
		std::cerr << ex.what() << std::endl;
		return 2;
	}
}
